from django.core.management import CommandError, call_command

from arkindex.documents.models import ElementPath
from arkindex.images.models import Image
from arkindex.project.tests import FixtureTestCase


class TestMoveLinesToParentCommand(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.single_page_type = cls.corpus.types.create(slug="single_page")
        cls.double_page_type = cls.corpus.types.create(slug="double_page")
        cls.paragraph_type = cls.corpus.types.create(slug="paragraph")
        cls.text_line_type = cls.corpus.types.get(slug="text_line")
        # Double page element
        cls.double_page = cls.corpus.elements.create(name="1", type=cls.double_page_type, polygon=[[0, 0], [0, 1000], [2000, 1000], [2000, 0], [0, 0]], image=Image.objects.first())
        # A single page element, child of the double page
        cls.single_page = cls.corpus.elements.create(name="1", type=cls.single_page_type, polygon=[[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]], image=Image.objects.first())
        cls.single_page.add_parent(cls.double_page)

    def test_merge_types_missing_single_page(self):
        with self.assertRaisesMessage(CommandError, 'Single page type 404 not found in Unit Tests'):
            call_command('move_lines_to_parent', corpus=self.corpus, single_page_type='404', double_page_type='double_page')

    def test_merge_types_missing_double_page(self):
        with self.assertRaisesMessage(CommandError, 'Double page type 404 not found in Unit Tests'):
            call_command('move_lines_to_parent', corpus=self.corpus, single_page_type='single_page', double_page_type='404')

    def test_move_lines_geographically_to_closest_parent(self):
        """This text_line overlaps enough with the given single page, it is moved
        """
        # Text line close to single page overlapping enough
        overlapping_text_line = self.corpus.elements.create(name="1", type=self.text_line_type, polygon=[[600, 0], [1100, 0], [600, 500], [1100, 500], [600, 0]], image=Image.objects.first())
        overlapping_text_line.add_parent(self.double_page)
        parent_paths = ElementPath.objects.filter(element_id=overlapping_text_line.id)
        self.assertEqual(len(parent_paths), 1)
        self.assertListEqual(list(parent_paths.values('path')), [{"path": [self.double_page.id]}])

        with self.assertNumQueries(23):
            call_command('move_lines_to_parent', corpus=self.corpus, single_page_type='single_page', double_page_type='double_page')

        # Make sure that lines have been moved
        parent_paths = ElementPath.objects.filter(element_id=overlapping_text_line.id)
        self.assertEqual(len(parent_paths), 1)
        self.assertListEqual(list(parent_paths.values('path')), [{"path": [self.double_page.id, self.single_page.id]}])

    def test_move_paragraph_geographically_to_closest_parent(self):
        """
        This paragraph overlaps enough with the given single page, it is moved
        """
        # Paragraph close to single page
        paragraph = self.corpus.elements.create(name="1", type=self.paragraph_type, polygon=[[10, 10], [10, 500], [500, 500], [500, 10], [10, 10]], image=Image.objects.first())
        paragraph.add_parent(self.double_page)
        parent_paths = ElementPath.objects.filter(element_id=paragraph.id)
        self.assertEqual(len(parent_paths), 1)
        self.assertListEqual(list(parent_paths.values('path')), [{"path": [self.double_page.id]}])

        with self.assertNumQueries(23):
            call_command('move_lines_to_parent', corpus=self.corpus, single_page_type='single_page', double_page_type='double_page')

        # Make sure that paragraph have been moved
        parent_paths = ElementPath.objects.filter(element_id=paragraph.id)
        self.assertEqual(len(parent_paths), 1)
        self.assertListEqual(list(parent_paths.values('path')), [{"path": [self.double_page.id, self.single_page.id]}])

    def test_do_not_move_line_not_close_enough(self):
        """
        This text line does not overlap enough with the single page, it stays under the double page
        """
        # Text line not close to single page not overlapping enough
        not_overlapping_text_line = self.corpus.elements.create(name="2", type=self.text_line_type, polygon=[[1500, 0], [1500, 100], [2000, 100], [2000, 0], [1500, 0]], image=Image.objects.first())
        not_overlapping_text_line.add_parent(self.double_page)
        parent_paths = ElementPath.objects.filter(element_id=not_overlapping_text_line.id)
        self.assertEqual(len(parent_paths), 1)
        self.assertListEqual(list(parent_paths.values('path')), [{"path": [self.double_page.id]}])

        with self.assertNumQueries(8):
            call_command('move_lines_to_parent', corpus=self.corpus, single_page_type='single_page', double_page_type='double_page')

        # Make sure that line has not been moved
        parent_paths = ElementPath.objects.filter(element_id=not_overlapping_text_line.id)
        self.assertEqual(len(parent_paths), 1)
        self.assertListEqual(list(parent_paths.values('path')), [{"path": [self.double_page.id]}])

    def test_move_line_weird_polygon_avoids_zero_division(self):
        """
        Some rotated rectangles may introduce some division by zero but should be supported
        they have either top_left.y == bottom_right.y or top_left.x == bottom_right.x
        These two values are subtracted and the result is used as quotient to compute the intersection
        This error should be avoided by computing the bounding rectangle
        """
        # Text line with top_left.y == bottom_right.y
        same_y_text_line = self.corpus.elements.create(name="1", type=self.text_line_type, polygon=[[10, 10], [20, 20], [520, 10], [500, 0], [10, 10]], image=Image.objects.first())
        # Text line with top_left.x == bottom_right.x
        same_x_text_line = self.corpus.elements.create(name="2", type=self.text_line_type, polygon=[[10, 10], [20, 20], [20, 520], [10, 500], [10, 10]], image=Image.objects.first())

        for line in [same_x_text_line, same_y_text_line]:
            line.add_parent(self.double_page)
            parent_paths = ElementPath.objects.filter(element_id=line.id)
            self.assertEqual(len(parent_paths), 1)
            self.assertListEqual(list(parent_paths.values('path')), [{"path": [self.double_page.id]}])

        with self.assertNumQueries(38):
            call_command('move_lines_to_parent', corpus=self.corpus, single_page_type='single_page', double_page_type='double_page')

        for line in [same_x_text_line, same_y_text_line]:
            # Make sure that line were not moved
            parent_paths = ElementPath.objects.filter(element_id=line.id)
            self.assertEqual(len(parent_paths), 1)
            self.assertListEqual(list(parent_paths.values('path')), [{"path": [self.double_page.id, self.single_page.id]}])
