from unittest.mock import call, patch

from arkindex.documents.tasks import move_selection
from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import User


class TestMoveSelection(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user2 = User.objects.create(email='user2@test.test', display_name='User 2', verified_email=True)
        cls.destination = cls.corpus.elements.get(name='Volume 2')
        cls.page_1 = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.page_2 = cls.corpus.elements.get(name='Volume 1, page 2r')
        cls.page_3 = cls.corpus.elements.get(name='Volume 1, page 1v')

    def test_no_rq_job(self):
        with self.assertRaises(AssertionError) as ctx:
            move_selection(corpus_id=self.corpus.id, destination=self.destination)
        self.assertEqual(str(ctx.exception), 'This task can only be run in a RQ job context.')

    @patch('arkindex.documents.tasks.get_current_job')
    def test_no_user_id(self, job_mock):
        job_mock.return_value.user_id = None
        with self.assertRaises(AssertionError) as ctx:
            move_selection(corpus_id=self.corpus.id, destination=self.destination)
        self.assertEqual(str(ctx.exception), 'This task requires a user ID to be defined on the RQ job.')

    @patch('arkindex.documents.tasks.get_current_job')
    @patch('arkindex.documents.tasks.move_element')
    def test_move_selected_elements(self, move_element_mock, job_mock):
        job_mock.return_value.user_id = self.user.id

        self.user.selected_elements.set([self.page_1, self.page_2])
        # Another user's selection should not be used
        self.user2.selected_elements.set([self.page_3])

        move_selection(corpus_id=self.corpus.id, destination=self.destination)

        self.assertEqual(job_mock.call_count, 1)
        self.assertEqual(job_mock().set_progress.call_count, 2)
        self.assertListEqual(job_mock().set_progress.call_args_list, [
            call(.0),
            call(.5),
        ])

        self.assertEqual(move_element_mock.call_count, 2)
        self.assertCountEqual(move_element_mock.call_args_list, [
            call(source=self.page_1, destination=self.destination),
            call(source=self.page_2, destination=self.destination)
        ])
