import uuid

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Classification, ClassificationState, Corpus, Element, MLClass
from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestClassifications(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.text = MLClass.objects.create(name='text', corpus=cls.corpus)
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.act_type = cls.corpus.types.get(slug='act')
        cls.element = Element.objects.get(name='Volume 1, page 1v')
        cls.folder = cls.corpus.elements.get(name='Volume 1')
        cls.worker_version_1 = WorkerVersion.objects.get(worker__slug='dla')
        cls.worker_version_2 = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_run_1 = cls.worker_version_1.worker_runs.get()
        cls.worker_run_2 = cls.worker_version_2.worker_runs.get()
        cls.internal_user = User.objects.get_by_natural_key('internal@internal.fr')

    def test_create_manual(self):
        """
        Creating a manual classification set auto fields correctly
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_version': '',
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()

        self.assertEqual(classification.worker_version, None)
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Validated)
        self.assertEqual(classification.confidence, 1)
        self.assertTrue(classification.high_confidence)
        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'element': str(self.element.id),
            'ml_class': str(self.text.id),
            'worker_run_id': None,
            'worker_version': None,
            'state': ClassificationState.Validated.value,
            'confidence': 1.0,
            'high_confidence': True,
        })

    def test_create_null_version(self):
        """
        A manual classification may be created specifying the version as null
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:classification-create'),
                data={
                    'element': str(self.element.id),
                    'ml_class': str(self.text.id),
                    'worker_version': None,
                },
                format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_manual_exists(self):
        """
        If a manual classification exists, creation must respond a 400_BAD_REQUEST
        with an explicit message
        """
        self.client.force_login(self.user)
        request = (
            reverse('api:classification-create'),
            {'element': str(self.element.id), 'ml_class': str(self.text.id)}
        )
        with self.assertNumQueries(8):
            response = self.client.post(*request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        with self.assertNumQueries(5):
            response = self.client.post(*request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'non_field_errors': ['A manual classification already exists for this element and this class.']
        })

    def test_create_worker_version_exists(self):
        """
        If a worker classification exists, creation must respond a 400_BAD_REQUEST
        with an explicit message
        """
        self.client.force_login(self.internal_user)
        request = (
            reverse('api:classification-create'),
            {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_version': str(self.worker_version_2.id),
                'confidence': '0.42',
                'high_confidence': False,
            }
        )
        response = self.client.post(*request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        with self.assertNumQueries(6):
            response = self.client.post(*request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['A classification of this worker version already exists for this element and this class.']
        })

    def test_create_writable_corpus(self):
        """
        Require the element and ML class to be in a writable corpus
        """
        self.client.force_login(self.user)
        ml_class = self.private_corpus.ml_classes.create(name='Heatmor')
        self.element.corpus = self.private_corpus
        self.element.save()

        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(ml_class.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            'element': [f'Invalid pk "{self.element.id}" - object does not exist.'],
            'ml_class': [f'Invalid pk "{ml_class.id}" - object does not exist.'],
        })

    def test_create_same_corpus(self):
        """
        Require the ML class to be in the same corpus as the element
        """
        self.client.force_login(self.user)
        self.private_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        ml_class = self.private_corpus.ml_classes.create(name='Heatmor')

        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(ml_class.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            'non_field_errors': ['Element and ML class are not in the same corpus']
        })

    def test_create_ignored_params(self):
        """
        Ensure confidence and high_confidence attributes cannot be changed on manual classifications
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'confidence': 0.5,
                'high_confidence': False,
                'state': 'rejected',
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()
        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'element': str(self.element.id),
            'ml_class': str(self.text.id),
            'worker_run_id': None,
            'worker_version': None,
            'state': ClassificationState.Validated.value,
            'confidence': 1,
            'high_confidence': True,
        })

    def test_create_without_permission(self):
        """
        Without permission, should not raise a 403 Forbidden: indicates that the resource exists
        Indicates that page PK does not exist instead
        """
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_worker_version_requires_internal(self):
        """
        Test creating a classification with a worker version requires the user to be internal
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_version': str(self.worker_version_1.id)
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'worker_version': ['An internal user is required to create a classification with a worker version.'],
            'confidence': ['This field is required to create a classification with a worker version.'],
            'high_confidence': ['This field is required to create a classification with a worker version.']
        })

    def test_create_worker_version_non_admin(self):
        """
        Test creating a classification with a worker version requires the user to be internal,
        and does not make an exception for admin users
        """
        self.assertFalse(self.superuser.is_internal)
        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_version': str(self.worker_version_1.id)
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'worker_version': ['An internal user is required to create a classification with a worker version.'],
            'confidence': ['This field is required to create a classification with a worker version.'],
            'high_confidence': ['This field is required to create a classification with a worker version.']
        })

    def test_create_worker_version(self):
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_version': str(self.worker_version_1.id),
                'confidence': 0.42,
                'high_confidence': False,
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()
        self.assertEqual(classification.worker_version, self.worker_version_1)
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Pending)
        self.assertEqual(classification.confidence, 0.42)
        self.assertFalse(classification.high_confidence)

    def test_create_worker_run_requires_internal(self):
        """
        Test creating a classification with a worker version requires the user to be internal
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_run_id': str(self.worker_run_1.id)
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'worker_run_id': ['An internal user is required to create a classification with a worker run.'],
            'confidence': ['This field is required to create a classification with a worker run.'],
            'high_confidence': ['This field is required to create a classification with a worker run.']
        })

    def test_create_worker_run_non_admin(self):
        """
        Test creating a classification with a worker run requires the user to be internal,
        and does not make an exception for admin users
        """
        self.assertFalse(self.superuser.is_internal)
        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_run_id': str(self.worker_run_1.id)
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'worker_run_id': ['An internal user is required to create a classification with a worker run.'],
            'confidence': ['This field is required to create a classification with a worker run.'],
            'high_confidence': ['This field is required to create a classification with a worker run.']
        })

    def test_create_worker_run(self):
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_run_id': str(self.worker_run_1.id),
                'confidence': 0.42,
                'high_confidence': False,
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()
        self.assertEqual(classification.worker_run, self.worker_run_1)
        self.assertEqual(classification.worker_version, self.worker_version_1)
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Pending)
        self.assertEqual(classification.confidence, 0.42)
        self.assertFalse(classification.high_confidence)

    def test_create_worker_version_xor_worker_run(self):
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_version': str(self.worker_version_1.id),
                'worker_run_id': str(self.worker_run_1.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            'non_field_errors': ['Only one of `worker_version` and `worker_run_id` may be set.']
        })

    def test_create_worker_run_not_found(self):
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(5):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_run_id': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            'worker_run_id': ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.'],
        })

    def test_create_zero_confidence(self):
        """
        Ensure CreateClassification accepts a confidence of 0
        """
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_version': str(self.worker_version_1.id),
                'confidence': 0,
                'high_confidence': False,
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()
        self.assertEqual(classification.worker_version, self.worker_version_1)
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Pending)
        self.assertEqual(classification.confidence, 0)
        self.assertFalse(classification.high_confidence)

    def test_create_manual_and_worker_version(self):
        """
        CreateClassification should allow creating the same ML class from a worker run,
        a worker version and no worker version on the same element.
        """
        self.client.force_login(self.internal_user)

        # Create a manual classification
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self.element.classifications.count(), 1)

        # Create a classification with the same class and a worker version
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_version': str(self.worker_version_1.id),
                'confidence': 0.5,
                'high_confidence': False,
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self.element.classifications.count(), 2)

        # Create a classification with the same class and a worker run
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:classification-create'), {
                'element': str(self.element.id),
                'ml_class': str(self.text.id),
                'worker_run_id': str(self.worker_run_1.id),
                'confidence': 0.5,
                'high_confidence': False,
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())

        self.assertEqual(self.element.classifications.count(), 3)

        self.assertCountEqual(
            list(self.element.classifications.values('worker_version_id', 'worker_run_id')),
            [
                {'worker_version_id': None, 'worker_run_id': None},
                {'worker_version_id': self.worker_version_1.id, 'worker_run_id': None},
                {'worker_version_id': self.worker_version_1.id, 'worker_run_id': self.worker_run_1.id},
            ]
        )

    def test_validate(self):
        classification = self.element.classifications.create(
            worker_version=self.worker_version_1,
            ml_class=self.text,
            confidence=.1
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.put(reverse('api:classification-validate', kwargs={'pk': classification.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'worker_version': str(self.worker_version_1.id),
            'worker_run': None,
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Validated.value,
            'confidence': classification.confidence,
            'high_confidence': False
        })

        # Ensure moderator has been set
        classification.refresh_from_db()
        self.assertEqual(classification.moderator, self.user)

    def test_validate_requires_login(self):
        classification = self.element.classifications.create(
            ml_class=self.text,
            confidence=.5,
        )
        with self.assertNumQueries(0):
            response = self.client.put(reverse('api:classification-validate', kwargs={'pk': classification.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_validate_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        classification = self.element.classifications.create(
            ml_class=self.text,
            confidence=.5,
        )
        with self.assertNumQueries(2):
            response = self.client.put(reverse('api:classification-validate', kwargs={'pk': classification.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_reject_worker_version(self):
        self.client.force_login(self.user)
        classification = self.element.classifications.create(
            worker_version=self.worker_version_1,
            ml_class=self.text,
            confidence=.1,
        )

        with self.assertNumQueries(7):
            response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'worker_version': str(self.worker_version_1.id),
            'worker_run': None,
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Rejected.value,
            'confidence': classification.confidence,
            'high_confidence': False
        })

        # Ensure moderator has been set
        classification.refresh_from_db()
        self.assertEqual(classification.moderator, self.user)

    def test_reject_manual_delete(self):
        """
        A classifications with no worker version should be deleted when rejected
        """
        self.client.force_login(self.user)
        classification = self.element.classifications.create(ml_class=self.text, confidence=.42)

        with self.assertNumQueries(6):
            response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(Classification.DoesNotExist):
            classification.refresh_from_db()

    def test_reject_requires_login(self):
        classification = self.element.classifications.create(ml_class=self.text, confidence=.42)
        with self.assertNumQueries(0):
            response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_reject_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        classification = self.element.classifications.create(ml_class=self.text, confidence=.42)
        with self.assertNumQueries(2):
            response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_repeatable_moderation(self):
        """
        A rejected classification can be re-validated
        """
        self.client.force_login(self.user)
        classification = self.element.classifications.create(
            ml_class=self.text,
            confidence=.5,
            moderator=self.user,
            state=ClassificationState.Validated,
            worker_version=self.worker_version_2,
        )

        # First try to reject
        with self.assertNumQueries(7):
            response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Rejected.value,
            'confidence': classification.confidence,
            'high_confidence': False,
            'worker_version': str(self.worker_version_2.id),
            'worker_run': None,
        })

        # Then try to validate
        with self.assertNumQueries(5):
            response = self.client.put(reverse('api:classification-validate', kwargs={'pk': classification.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Validated.value,
            'confidence': classification.confidence,
            'high_confidence': False,
            'worker_version': str(self.worker_version_2.id),
            'worker_run': None,
        })

    def test_create_selection_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': self.corpus.id, 'ml_class': self.text.id, 'mode': 'create'}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={'selection': False})
    def test_create_selection_no_selection(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': self.corpus.id, 'ml_class': self.text.id, 'mode': 'create'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'non_field_errors': ['Selection is not available on this instance.']})

    def test_create_selection_private_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': self.private_corpus.id, 'ml_class': self.text.id, 'mode': 'create'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_selection_writable_corpus(self):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': self.corpus.id, 'ml_class': self.text.id, 'mode': 'create'}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {
            'detail': 'You do not have contributor access to this corpus.',
        })

    def test_create_selection_non_existing_corpus(self):
        random_uuid = uuid.uuid4()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': random_uuid, 'ml_class': self.text.id, 'mode': 'create'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            'corpus_id': [f'Invalid pk "{random_uuid}" - object does not exist.'],
        })

    def test_create_selection_wrong_corpus(self):
        line = MLClass.objects.create(name='line', corpus=self.private_corpus)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': self.corpus.id, 'ml_class': line.id, 'mode': 'create'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            'non_field_errors': [f'Ml class {line} does not belong to the corpus {self.corpus}']
        })

    def test_create_selection_wrong_ml_class(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': self.corpus.id, 'ml_class': 'ffffffff-ffff-ffff-ffff-ffffffffffff', 'mode': 'create'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            'ml_class': ['Ml class ffffffff-ffff-ffff-ffff-ffffffffffff not found']
        })

    def test_create_selection_wrong_mode(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': self.corpus.id, 'ml_class': self.text.id, 'mode': 'test'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_selection_create(self):
        act_x = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name='Act xxx',
        )
        act_y = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name='Act yyy',
        )

        self.assertFalse(self.folder.classifications.exists())
        self.assertFalse(self.element.classifications.exists())

        self.client.force_login(self.superuser)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:elements-selection'),
                data={'ids': [self.folder.id, self.element.id, act_x.id, act_y.id]},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        with self.assertNumQueries(4):
            response = self.client.get(reverse('api:elements-selection'))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        results = response.json()['results']
        self.assertEqual(len(results), 4)

        self.assertCountEqual(
            [element['id'] for element in results],
            [str(self.element.id), str(self.folder.id), str(act_x.id), str(act_y.id)],
        )

        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': self.corpus.id, 'ml_class': self.text.id, 'mode': 'create'}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertTrue(self.folder.classifications.exists())
        classification = self.folder.classifications.get()
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Validated)
        self.assertIsNone(classification.worker_version)
        self.assertEqual(classification.confidence, 1)

        self.assertTrue(self.element.classifications.exists())
        classification = self.element.classifications.get()
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Validated)
        self.assertIsNone(classification.worker_version)
        self.assertEqual(classification.confidence, 1)

    def test_create_selection_validate(self):
        line = MLClass.objects.create(name='line', corpus=self.private_corpus)
        act_x = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name='Act xxx',
        )
        Classification.objects.create(
            element=act_x,
            worker_version=self.worker_version_1,
            state=ClassificationState.Pending,
            high_confidence=True,
            ml_class=line
        )

        for e in [self.folder, self.element]:
            Classification.objects.create(
                element=e,
                worker_version=self.worker_version_1,
                state=ClassificationState.Pending,
                high_confidence=True,
                ml_class=self.text
            )
            Classification.objects.create(
                element=e,
                worker_version=self.worker_version_2,
                state=ClassificationState.Pending,
                high_confidence=False,
                ml_class=self.text
            )

        self.client.force_login(self.superuser)
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.folder.id, self.element.id, act_x.id]},
            format='json'
        )

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': self.corpus.id, 'mode': 'validate'}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        for e in [self.folder, self.element]:
            classification = e.classifications.get(state=ClassificationState.Validated)
            self.assertTrue(classification.high_confidence)
            self.assertEqual(classification.worker_version, self.worker_version_1)

            classification = e.classifications.get(state=ClassificationState.Pending)
            self.assertFalse(classification.high_confidence)
            self.assertEqual(classification.worker_version, self.worker_version_2)

        classification = act_x.classifications.get()
        self.assertEqual(classification.state, ClassificationState.Pending)

    def test_create_selection_create_requires_ml_class(self):
        self.client.force_login(self.superuser)
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.folder.id, self.element.id]},
            format='json'
        )

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse('api:classification-selection'),
                data={'corpus_id': self.corpus.id, 'mode': 'create'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {'ml_class': ['ML class is required when adding a classification']})
