from unittest.mock import call, patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestDestroyWorkerResults(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.version = WorkerVersion.objects.get(worker__slug='reco')
        cls.page = cls.corpus.elements.get(name='Volume 1, page 2r')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_wrong_acl(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.private_corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have admin access to this corpus.'}
        )

    def test_wrong_corpus_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': '12341234-1234-1234-1234-123412341234'}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(
            response.json(),
            {'detail': 'Not found.'}
        )

    @patch('arkindex.project.triggers.documents_tasks.worker_results_delete.delay')
    def test_no_filter(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            version_id=None,
            element_id=None,
            user_id=self.user.id,
            description='Worker results deletion',
        ))

    @patch('arkindex.project.triggers.documents_tasks.worker_results_delete.delay')
    def test_filter_version(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + f'?worker_version_id={self.version.id}',
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            version_id=self.version.id,
            element_id=None,
            user_id=self.user.id,
            description=f"Deletion of results produced by worker version: {self.version}"
        ))

    @patch('arkindex.project.triggers.documents_tasks.worker_results_delete.delay')
    def test_filter_element(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + f'?element_id={self.page.id}'
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            version_id=None,
            element_id=self.page.id,
            user_id=self.user.id,
            description='Worker results deletion',
        ))

    @patch('arkindex.project.triggers.documents_tasks.worker_results_delete.delay')
    def test_filter_version_element(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + f'?worker_version_id={self.version.id}&element_id={self.page.id}'
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            version_id=self.version.id,
            element_id=self.page.id,
            user_id=self.user.id,
            description=f"Deletion of results produced by worker version: {self.version}"
        ))

    def test_invalid_version_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + '?worker_version_id=lol'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'worker_version_id': ['Invalid UUID.']},
        )

    def test_wrong_version_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + '?worker_version_id=12341234-1234-1234-1234-123412341234'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'worker_version_id': ['This worker version does not exist.']}
        )

    def test_invalid_element_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + '?element_id=lol'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'element_id': ['Invalid UUID.']}
        )

    def test_wrong_element_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + '?element_id=12341234-1234-1234-1234-123412341234'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'element_id': ['This element does not exist in the specified corpus.']}
        )

    @override_settings(ARKINDEX_FEATURES={'selection': False})
    def test_selection_feature_flag(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + '?use_selection=true'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'use_selection': ['Selection is not available on this instance.']}
        )

    @override_settings(ARKINDEX_FEATURES={'selection': True})
    def test_selection_no_element_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + f'?use_selection=true&element_id={self.page.id}'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'use_selection': [
                'use_selection and element_id cannot be used simultaneously.',
                'No elements of the specified corpus have been selected.'
            ]}
        )

    @override_settings(ARKINDEX_FEATURES={'selection': True})
    def test_selection_empty(self):
        self.client.force_login(self.user)
        self.assertFalse(self.user.selected_elements.exists())
        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + '?use_selection=true'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'use_selection': ['No elements of the specified corpus have been selected.']}
        )

    @override_settings(ARKINDEX_FEATURES={'selection': True})
    def test_selection_empty_for_corpus(self):
        self.superuser.selected_elements.add(self.page)
        self.assertFalse(self.user.selected_elements.filter(corpus=self.private_corpus).exists())
        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.private_corpus.id)})
                + '?use_selection=true'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(
                response.json(),
                {'use_selection': ['No elements of the specified corpus have been selected.']}
            )

    @override_settings(ARKINDEX_FEATURES={'selection': True})
    @patch('arkindex.project.triggers.documents_tasks.selection_worker_results_delete.delay')
    def test_selection(self, delay_mock):
        self.user.selected_elements.add(self.page)
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + '?use_selection=true'
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            version_id=None,
            user_id=self.user.id,
            description=f"Deletion of worker results on selected elements in {self.corpus.name}"
        ))

    @override_settings(ARKINDEX_FEATURES={'selection': True})
    @patch('arkindex.project.triggers.documents_tasks.selection_worker_results_delete.delay')
    def test_selection_version_filter(self, delay_mock):
        self.user.selected_elements.add(self.page)
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.delete(
                reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id)})
                + f'?use_selection=true&worker_version_id={self.version.id}'
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            version_id=self.version.id,
            user_id=self.user.id,
            description=f"Deletion of worker results on selected elements in {self.corpus.name} "
                        f"produced by {self.version}"
        ))
