import hashlib
import uuid
from collections import defaultdict
from textwrap import dedent

from django.db.models import Count, Max, Q
from django.shortcuts import get_object_or_404, redirect
from django.utils import timezone
from drf_spectacular.utils import OpenApiExample, extend_schema, extend_schema_view
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveUpdateAPIView,
    UpdateAPIView,
)
from rest_framework.permissions import SAFE_METHODS
from rest_framework.response import Response
from rest_framework.views import APIView

from arkindex.ponos.authentication import AgentAuthentication
from arkindex.ponos.keys import load_private_key
from arkindex.ponos.models import Agent, Artifact, Farm, Secret, State, Task, Workflow
from arkindex.ponos.permissions import (
    IsAgent,
    IsAgentOrArtifactAdmin,
    IsAgentOrInternal,
    IsAgentOrTaskAdmin,
    IsAssignedAgentOrReadOnly,
)
from arkindex.ponos.recipe import parse_recipe
from arkindex.ponos.renderers import PublicKeyPEMRenderer
from arkindex.ponos.serializers import (
    AgentActionsSerializer,
    AgentCreateSerializer,
    AgentDetailsSerializer,
    AgentLightSerializer,
    AgentStateSerializer,
    ArtifactSerializer,
    ClearTextSecretSerializer,
    FarmSerializer,
    NewTaskSerializer,
    TaskDefinitionSerializer,
    TaskSerializer,
    TaskTinySerializer,
    WorkflowSerializer,
)
from arkindex.process.models import Process
from arkindex.project.mixins import ProcessACLMixin
from arkindex.project.permissions import IsAuthenticated, IsVerified
from arkindex.users.models import Role
from rest_framework_simplejwt.views import TokenRefreshView


class PublicKeyEndpoint(APIView):
    """
    Fetch the server's public key in PEM format to perform agent registration.
    """

    renderer_classes = (PublicKeyPEMRenderer,)

    @extend_schema(
        operation_id="GetPublicKey",
        responses={200: {"type": "string"}},
        tags=["ponos"],
        auth=[{"agentAuth": []}],
        examples=[
            OpenApiExample(
                name="Public key response",
                response_only=True,
                media_type="application/x-pem-file",
                status_codes=["200"],
                value=dedent(
                    """
            -----BEGIN PUBLIC KEY-----
            MHYwEAYHKoZIzj0CAQYFK4EEACIDYgAEmK2L6lwGzSVZwFSo0eR1z4XV6jJwjeWK
            YCiPKdMcQnn6u5J016k9U8xZm6XyFnmgvkhnC3wreGBTFzwLCLZCD+F3vo5x8ivz
            aTgNWsA3WFlqjSIEGz+PAVHSNMobBaJm
            -----END PUBLIC KEY-----
            """
                ),
            )
        ],
    )
    def get(self, request, *args, **kwargs):
        return Response(load_private_key().public_key())


@extend_schema(tags=["ponos"])
@extend_schema_view(
    get=extend_schema(description=dedent("""
        Retrieve a Ponos workflow status.

        Requires **guest** access to its process.
    """).strip()),
    put=extend_schema(description=dedent("""
        Update a workflow's status and tasks.

        Requires **admin** access to its process.
    """).strip()),
    patch=extend_schema(description=dedent("""
        Partially update a workflow's status and tasks.

        Requires **admin** access to its process.
    """).strip()),
)
class WorkflowDetails(ProcessACLMixin, RetrieveUpdateAPIView):
    """
    Retrieve information about a workflow, or update its state.
    Updating a workflow's state to :attr:`~arkindex.ponos.models.State.Stopping` will cause it to stop.
    """

    permission_classes = (IsVerified, )
    queryset = Workflow.objects.prefetch_related("tasks__parents").annotate(
        last_run=Max("tasks__run")
    )
    serializer_class = WorkflowSerializer

    def perform_update(self, serializer):
        serializer.instance.stop()

    def check_object_permissions(self, request, workflow):
        super().check_object_permissions(request, workflow)
        if self.request.user.is_admin:
            return

        required_access = Role.Admin.value
        # Allow a user with a read access on the process to retrieve its workflow
        if request.method in SAFE_METHODS:
            required_access = Role.Guest.value

        process = get_object_or_404(
            Process,
            workflow_id=workflow.id
        )
        access_level = self.process_access_level(process)
        if not access_level:
            raise NotFound
        if access_level < required_access:
            raise PermissionDenied(detail='You do not have an admin access to this process.')


@extend_schema(tags=["ponos"])
@extend_schema_view(
    get=extend_schema(
        operation_id="RetrieveTaskFromAgent", description="Retrieve a Ponos task status"
    ),
    put=extend_schema(
        operation_id="UpdateTaskFromAgent", description="Update a task, from an agent"
    ),
    patch=extend_schema(
        operation_id="PartialUpdateTaskFromAgent",
        description="Partially update a task, from an agent",
    ),
)
class TaskDetailsFromAgent(RetrieveUpdateAPIView):
    """
    Retrieve information about a single task, including its logs.
    Authenticated agents assigned to a task can use this endpoint to report its state.
    """

    # Avoid stale read when a recently assigned agent wants to update
    # the state of one of its tasks
    queryset = Task.objects.all().using("default")
    permission_classes = (IsAssignedAgentOrReadOnly,)
    serializer_class = TaskSerializer


@extend_schema_view(
    post=extend_schema(
        operation_id="CreateAgent",
        tags=["ponos"],
    ),
)
class AgentRegister(CreateAPIView):
    """
    Perform agent registration and authentication.
    """

    serializer_class = AgentCreateSerializer

    def get_object(self):
        if not hasattr(self.request, "data") or "public_key" not in self.request.data:
            return
        key_hash = uuid.UUID(
            hashlib.md5(self.request.data["public_key"].encode("utf-8")).hexdigest()
        )
        return Agent.objects.filter(id=key_hash).first()

    def get_serializer(self, *args, **kwargs):
        return super().get_serializer(self.get_object(), *args, **kwargs)


@extend_schema_view(
    post=extend_schema(
        operation_id="RefreshAgentToken",
        tags=["ponos"],
    )
)
class AgentTokenRefresh(TokenRefreshView):
    """
    Refresh a Ponos agent token when it expires
    """


@extend_schema(tags=["ponos"])
class AgentDetails(RetrieveAPIView):
    """
    Retrieve details of an agent including its running tasks

    Requires authentication with a verified e-mail address.
    """
    permission_classes = (IsVerified, )
    serializer_class = AgentDetailsSerializer
    queryset = Agent.objects.all()


@extend_schema(
    description="List the state of all Ponos agents",
    tags=["ponos"],
)
class AgentsState(ListAPIView):
    """
    List all agents on the system with their health state.

    Requires authentication with a verified e-mail address.
    """
    permission_classes = (IsVerified, )
    serializer_class = AgentStateSerializer

    queryset = (
        Agent.objects.all()
        .annotate(
            running_tasks_count=Count("tasks", filter=Q(tasks__state=State.Running))
        )
        .prefetch_related("farm")
        .order_by("hostname")
    )


@extend_schema(tags=["ponos"])
class AgentActions(RetrieveAPIView):
    """
    Fetch the next actions an agent should perform.
    """

    permission_classes = (IsAgent,)
    authentication_classes = (AgentAuthentication,)
    serializer_class = AgentActionsSerializer

    def get_object(self):
        return self.request.user

    def retrieve(self, request, *args, **kwargs):
        # Update agent load and last_ping timestamp
        errors = defaultdict(list)
        cpu_load = request.query_params.get("cpu_load")
        ram_load = request.query_params.get("ram_load")
        if not cpu_load:
            errors["cpu_load"].append("This query parameter is required.")
        if not ram_load:
            errors["ram_load"].append("This query parameter is required.")
        if errors:
            raise ValidationError(errors)
        # Handle fields validation with DRF as for a PATCH
        agent_serializer = AgentLightSerializer(
            self.request.user,
            data={"cpu_load": cpu_load, "ram_load": ram_load},
            partial=True,
        )
        agent_serializer.is_valid(raise_exception=True)
        # Update agent load and last_ping attributes
        agent_serializer.save(last_ping=timezone.now())
        # Retrieve next tasks after the DB has been updated with the new agent load
        response = super().retrieve(self, request, *args, **kwargs)
        return response


@extend_schema_view(
    get=extend_schema(
        operation_id="RetrieveTaskDefinition",
        tags=["ponos"],
    )
)
class TaskDefinition(RetrieveAPIView):
    """
    Obtain a task's definition as an agent or admin.
    This holds all the required data to start a task, except for the artifacts.
    """

    # We need to specify the default database to avoid stale reads
    # when a task is updated by an agent, then the agent immediately fetches its definition
    queryset = Task.objects.using("default").select_related("workflow")
    permission_classes = (IsAgent,)
    serializer_class = TaskDefinitionSerializer


@extend_schema(tags=["ponos"])
@extend_schema_view(
    get=extend_schema(
        operation_id="ListArtifacts",
        description="List all the artifacts of a task",
    ),
    post=extend_schema(
        operation_id="CreateArtifact", description="Create an artifact on a task"
    ),
)
class TaskArtifacts(ListCreateAPIView):
    """
    List all artifacts linked to a task or create one
    """

    # Used for OpenAPI schema serialization: the ID in the path is the task ID
    queryset = Task.objects.none()
    permission_classes = (IsAuthenticated, )
    serializer_class = ArtifactSerializer

    # Force no pagination, even when global settings add them
    pagination_class = None

    def get_task(self):
        if "pk" not in self.kwargs:
            return Artifact.objects.none()
        return get_object_or_404(Task, pk=self.kwargs["pk"])

    def get_queryset(self):
        task = self.get_task()
        return task.artifacts.all()

    def perform_create(self, serializer):
        user = self.request.user
        if not user.is_staff and not getattr(user, "is_agent", False):
            raise PermissionDenied()

        # Assign task when creating through the API
        serializer.save(task=self.get_task())


class TaskArtifactDownload(APIView):
    """
    Redirects to the S3 URL of an artifact in order to download it.
    """

    permission_classes = (IsAgentOrArtifactAdmin, )

    def get_object(self, pk, path):
        artifact = get_object_or_404(Artifact, task_id=pk, path=path)
        self.check_object_permissions(self.request, artifact)
        return artifact

    @extend_schema(
        operation_id="DownloadArtifact",
        responses={302: None},
        tags=["ponos"],
    )
    def get(self, request, *args, **kwargs):
        artifact = self.get_object(*args, **kwargs)
        return redirect(artifact.s3_get_url)


@extend_schema_view(
    post=extend_schema(
        operation_id="CreateTask",
        tags=["ponos"],
    )
)
class TaskCreate(CreateAPIView):
    """
    Create a task with a parent
    """

    serializer_class = NewTaskSerializer

    def perform_create(self, serializer):
        """
        Merge workflow's recipe environment variables with the ones set on the task
        Task variables have priority over Workflow variables
        """
        env, _ = parse_recipe(
            Workflow.objects.get(id=serializer.validated_data["workflow_id"]).recipe
        )
        env.update(serializer.validated_data["env"])
        serializer.save(env=env)


@extend_schema(tags=["ponos"])
@extend_schema_view(
    put=extend_schema(
        description="Update a task, allowing humans to change the task's state"
    ),
    patch=extend_schema(
        description="Partially update a task, allowing humans to change the task's state"
    ),
)
class TaskUpdate(UpdateAPIView):
    """
    Admins and task creators can use this endpoint to update a task
    """
    permission_classes = (IsAgentOrTaskAdmin, )
    queryset = Task.objects.all()
    serializer_class = TaskTinySerializer


@extend_schema_view(
    get=extend_schema(
        operation_id="RetrieveSecret",
        tags=["ponos"],
    )
)
class SecretDetails(RetrieveAPIView):
    """
    Retrieve a Ponos secret content as cleartext
    """

    permission_classes = (IsAgentOrInternal, )
    serializer_class = ClearTextSecretSerializer

    def get_object(self):
        return get_object_or_404(Secret, name=self.kwargs.get("name"))


@extend_schema(tags=["ponos"])
class FarmList(ListAPIView):
    """
    List all available farms
    """

    serializer_class = FarmSerializer
    queryset = Farm.objects.order_by("name")
