import logging
import os.path
import random
import uuid
from collections import namedtuple
from datetime import timedelta
from hashlib import sha256

from botocore.exceptions import ClientError
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import MinLengthValidator, MinValueValidator, RegexValidator
from django.db import models, transaction
from django.db.models import Q
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from enumfields import Enum, EnumField
from yaml import YAMLError

from arkindex.ponos.aws import object_url, s3
from arkindex.ponos.fields import CommaSeparatedListField, StringDictField
from arkindex.ponos.keys import gen_nonce
from arkindex.ponos.managers import TaskManager
from arkindex.ponos.recipe import parse_recipe, recipe_depth
from arkindex.ponos.validators import MaxValueValidator
from rest_framework_simplejwt.tokens import RefreshToken

# Maximum allowed time until an agent is considered inactive since last request
AGENT_TIMEOUT = timedelta(
    **getattr(settings, "PONOS_ACTIVE_AGENT_TIMEOUT", {"seconds": 30})
)
# Estimation of required resources to run a task on an agent. Defaults to 1 core and 1GB of RAM
AGENT_SLOT = getattr(settings, "PONOS_TASK_SLOT", {"cpu": 1, "ram": 1e9})

Action = namedtuple("Action", "action, task")
Action.__doc__ = """
Describes an instruction sent to an agent.
Can optionally be associated with a :class:`Task`.
"""

Action.action.__doc__ = """
Type of action to perform.

:type: arkindex.ponos.models.ActionType
"""

Action.task.__doc__ = """
Optionally associated Task, when relevant.

:type: arkindex.ponos.models.Task
"""

logger = logging.getLogger(__name__)


class ActionType(Enum):
    """
    Describes which action an agent should perform.
    """

    StartTask = "start_task"
    """
    Instruct the agent to start a new task.
    The action must have an associated task.
    """

    StopTask = "stop_task"
    """
    Instruct the agent to stop a running task.
    The action must have an associated task.
    """

    Kill = "kill"
    """
    Instruct the agent to shut itself down.
    This is the only case where the agent will have a 0 exit code.
    """


def generate_seed() -> str:
    return "{:064x}".format(random.getrandbits(256))


class Farm(models.Model):
    """
    A group of agents, whose ID and seed can be used to register new agents
    automatically in auto-scaling contexts.
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    name = models.CharField(max_length=250)
    seed = models.CharField(
        max_length=64,
        unique=True,
        default=generate_seed,
        validators=[RegexValidator(r"^[0-9a-f]{64}$")],
    )

    def __str__(self) -> str:
        return "Farm {}".format(self.name)


class Agent(models.Model):
    """
    A remote host that can run tasks.
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    farm = models.ForeignKey(Farm, on_delete=models.PROTECT)
    public_key = models.TextField()

    hostname = models.SlugField(max_length=64, db_index=False)
    cpu_cores = models.PositiveSmallIntegerField(validators=[MinValueValidator(1)])
    cpu_frequency = models.BigIntegerField(validators=[MinValueValidator(1)])
    # Total amount of RAM on this agent in bytes
    ram_total = models.BigIntegerField(validators=[MinValueValidator(1)])
    # Last minute average CPU load measure on this agent
    cpu_load = models.FloatField(null=True, blank=True)
    # Last RAM load measure expressed as a percentage (0 ≤ ram_load ≤ 1)
    ram_load = models.FloatField(null=True, blank=True)
    last_ping = models.DateTimeField(editable=False)

    @property
    def token(self) -> RefreshToken:
        """
        JSON Web Token for this agent.
        """
        return RefreshToken.for_user(self)

    @property
    def active(self) -> bool:
        return self.last_ping >= timezone.now() - AGENT_TIMEOUT

    def __str__(self) -> str:
        return self.hostname

    def delete(self) -> None:
        if self.tasks.exclude(state__in=FINAL_STATES).exists():
            raise ValidationError(
                "This agent has one or more tasks in non-final states."
            )
        return super().delete()

    def _estimate_new_tasks_cost(self, tasks=1):
        """
        Metric used to estimate the load on this agent starting new tasks.
        Used as the cost function to minimize overall agents load while attributing tasks.

        :param tasks: Number of tasks to estimate the cost for.
        :returns: A cost expressed as a percentage. If > 1, the agent would be overloaded.
        """
        cpu_cost = (self.cpu_load + tasks * AGENT_SLOT["cpu"]) / self.cpu_cores
        ram_cost = self.ram_load + tasks * AGENT_SLOT["ram"] / self.ram_total
        return max(cpu_cost, ram_cost)

    def next_tasks(self):
        """
        Compute the next tasks that should be run by an agent.

        :returns: A list of tasks.
        """
        if self._estimate_new_tasks_cost() >= 1:
            # The capacity of this agent does not allow a new task to be started
            return []

        # Filter pending task on the agent farm ordered by higher priority first and then seniority (older first)
        pending_tasks = Task.objects.filter(
            Q(workflow__farm_id=self.farm_id)
            & Q(
                Q(state=State.Pending, agent=None) | Q(state=State.Unscheduled, depth=0)
            )
        ).order_by("-priority", "updated")

        if not pending_tasks.exists():
            return []

        # Retrieve active agents within the same farm
        active_agents = Agent.objects.filter(
            last_ping__gte=timezone.now() - AGENT_TIMEOUT
        ).filter(farm_id=self.farm_id)

        # List available gpus (without any pending or running task assigned)
        available_gpus = list(
            self.gpus.filter(Q(task__isnull=True) & ~Q(task__state__in=ACTIVE_STATES))
        )

        # Simulate the attribution of pending tasks minimizing overall load
        attributed_tasks = {agent: [] for agent in active_agents}
        for task in pending_tasks:

            # High priority for tasks with GPU requirements
            if task.requires_gpu:
                if available_gpus:
                    task.gpu = available_gpus.pop()
                    min_cost_agent = self
                else:
                    # Skip tasks requiring GPU when none is available
                    logger.info(f"No GPU available for task {task.id} - {task}")
                    continue
            else:
                # Compare the cost of adding a new task on compatible agents
                min_cost_agent = min(
                    active_agents,
                    key=lambda agent: agent._estimate_new_tasks_cost(
                        tasks=len(attributed_tasks[agent]) + 1
                    ),
                )

            # Append the task to the queue of the agent with the minimal cost
            tasks = attributed_tasks[min_cost_agent]
            if min_cost_agent._estimate_new_tasks_cost(len(tasks) + 1) > 1:
                # Attributing the next task would overload the system
                break
            tasks.append(task)

        # Return tasks attributed to the agent making the request
        return attributed_tasks[self]

    @transaction.atomic
    def next_actions(self):
        """
        Compute the next actions to send to an agent.

        This method must run in a single transaction to avoid fetching tasks that are being
        assigned to another agent.

        :returns: List of :obj:`Action`.
        """
        pending_tasks = self.tasks.filter(state=State.Pending)
        next_tasks = self.next_tasks()
        stop_tasks = self.tasks.filter(state=State.Stopping)

        actions = []
        for task in stop_tasks:
            actions.append(Action(ActionType.StopTask, task))
        for task in pending_tasks:
            actions.append(Action(ActionType.StartTask, task))
        for task in next_tasks:
            task.agent = self
            task.state = State.Pending
            task.save()
            actions.append(Action(ActionType.StartTask, task))
        return actions


class State(Enum):
    """
    Describes the possible :class:`Task` states.

    Inspired by
    `Taskcluster's states <https://docs.taskcluster.net/docs/manual/tasks/runs>`_
    """

    Unscheduled = "unscheduled"
    """
    This is the default initial state of all tasks. Tasks will keep this state in two cases:

    * They do not have dependencies in a :class:`Workflow`, but no :class:`Agent`
      has yet been assigned to them;
    * They depend on other tasks that are not yet :attr:`~State.Completed`.
    """

    Pending = "pending"
    """
    Intermediate state where a task is being assigned to an :class:`Agent` and starting.

    Tasks without dependencies will enter this state when they are assigned to an agent.
    Tasks with dependencies will reach this state as soon as all their dependencies are met.

    Once a task is assigned to an agent, the agent is responsible for updating its state.
    """

    Running = "running"
    """
    State where a task is currently being run by an :class:`Agent`.
    Agents will update the state of their assigned pending tasks to Running as soon
    as the Docker container has started. API users may start monitoring for logs.
    """

    Completed = "completed"
    """
    State where a task has finished successfully.
    Reaching this state means the artifacts are available, and any dependent tasks with
    all their dependencies met will enter the Pending state.
    """

    Failed = "failed"
    """
    State where a task has been run, but ended with a non-zero error code.
    More information might be found in the task's logs.
    """

    Error = "error"
    """
    State where a task may or may not have been run, but an error occurred on the
    :class:`Agent`'s side. This is usually caused by Docker or host-related issues.
    """

    Stopping = "stopping"
    """
    State that can be set by API users to ask for a running task to be stopped.
    The task's assigned agent will be instructed to stop the task, and will report the
    Stopped state when done.
    """

    Stopped = "stopped"
    """
    State where a task that entered the Stopping state has successfully stopped.
    """


# States where a task is considered final.
# Once a task reaches a final state, its state should no longer change.
FINAL_STATES = (
    State.Completed,
    State.Failed,
    State.Error,
    State.Stopped,
)

# States where a task is considered active on an agent
ACTIVE_STATES = (
    State.Pending,
    State.Running,
    State.Stopping,
)

# Tasks priority to determine the overall state of multiple tasks
# If there are failed tasks, the workflow is failed.
# Else, if there are any errors, the workflow errored.
# If there are any running tasks, no matter in which state another task may be,
# the workflow is set as running to prevent retrying and allow stopping.
STATES_ORDERING = [
    State.Running,
    State.Failed,
    State.Error,
    State.Stopping,
    State.Stopped,
    State.Pending,
    State.Unscheduled,
    State.Completed,
]


def recipe_validator(value) -> None:
    """
    Validator for the ``recipe`` field in a :class:`Workflow`.

    :raises ValidationError: When the recipe is not valid YAML, or not a valid recipe.
    """
    try:
        parse_recipe(value)
    except (YAMLError, AssertionError) as e:
        raise ValidationError(str(e))


class Workflow(models.Model):
    """
    A group of tasks that can depend on each other,
    created from a YAML recipe.
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    recipe = models.TextField(validators=[recipe_validator])
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    finished = models.DateTimeField(blank=True, null=True)
    farm = models.ForeignKey(
        to="ponos.Farm", related_name="workflows", on_delete=models.PROTECT
    )

    class Meta:
        ordering = ("-updated",)
        constraints = [
            # A workflow cannot be finished before it is created
            models.CheckConstraint(
                check=models.Q(finished=None)
                | models.Q(finished__gte=models.F("created")),
                name="ponos_workflow_finished_after_created",
            )
        ]

    def __str__(self) -> str:
        return str(self.id)

    @cached_property
    def recipes(self):
        _, recipes = parse_recipe(self.recipe)
        return recipes

    def get_absolute_url(self) -> str:
        """
        :returns: URL to the workflow details API for this workflow.
        """
        return reverse("ponos:workflow-details", args=[self.id])

    @transaction.atomic
    def build_tasks(self, run: int = 0):
        """
        Parse this workflow's recipe and create unscheduled :class:`Task` instances in database.

        :param run: Run number to assign to each task.
        :returns: A dict mapping task slugs to task instances.
        :rtype: dict
        """
        assert isinstance(run, int)
        assert run >= 0
        assert not self.tasks.filter(run=run).exists(), "Run {} already exists".format(
            run
        )

        # Create tasks without any parent
        tasks = {
            slug: self.tasks.create(
                run=run,
                slug=slug,
                tags=recipe.tags,
                depth=recipe_depth(slug, self.recipes),
                image=recipe.image,
                command=recipe.command,
                shm_size=recipe.shm_size,
                env=recipe.environment,
                has_docker_socket=recipe.has_docker_socket,
                image_artifact=Artifact.objects.get(id=recipe.artifact)
                if recipe.artifact
                else None,
                requires_gpu=recipe.requires_gpu,
                extra_files=recipe.extra_files if recipe.extra_files else {},
            )
            for slug, recipe in self.recipes.items()
        }

        # Apply parents
        for slug, recipe in self.recipes.items():
            task = tasks[slug]
            task.parents.set([tasks[parent] for parent in recipe.parents])

        return tasks

    @property
    def state(self) -> State:
        """
        Deduce the workflow's state from the tasks of its latest run.
        A workflow's state is deduced by picking the first state that any tasks in this run have, in this order:

        #. :attr:`~State.Running`
        #. :attr:`~State.Failed`
        #. :attr:`~State.Error`
        #. :attr:`~State.Stopping`
        #. :attr:`~State.Stopped`
        #. :attr:`~State.Pending`
        #. :attr:`~State.Unscheduled`
        """
        return self.get_state(self.get_last_run())

    def get_state(self, run):
        """
        A separate method to get a workflow's state on a given run.

        Most users will only use the :meth:`Workflow.state` property to get the state for a workflow's last run.
        However, when trying to get a state for many workflows at once, using ``.annotate(last_run=Max('tasks__run'))``
        and using the annotation with this method will prevent many useless SQL requests.

        Further performance improvements can be achieved with ``prefetch_related('tasks')``.
        """
        # Negative run numbers never have tasks
        if run < 0:
            return State.Unscheduled

        # This prevents performing another SQL request when tasks have already been prefetched.
        # See https://stackoverflow.com/a/19651840/5990435
        if (
            hasattr(self, "_prefetched_objects_cache")
            and self.tasks.field.remote_field.get_cache_name()
            in self._prefetched_objects_cache
        ):
            task_states = set(t.state for t in self.tasks.all() if t.run == run)
        else:
            task_states = set(
                self.tasks.filter(run=run).values_list("state", flat=True)
            )

        # This run has no tasks
        if not task_states:
            return State.Unscheduled

        # All tasks have the same state
        if len(task_states) == 1:
            return task_states.pop()

        for state in STATES_ORDERING:
            if state in task_states:
                return state

        raise NotImplementedError("Something went wrong")

    def get_last_run(self) -> int:
        """
        Get the last run number. If the ``last_run`` attribute is defined on this workflow,
        possibly from a ``.annotate(last_run=Max('tasks__run'))`` annotation in a Django QuerySet,
        this method will return the attribute's value instead of making another SQL request.
        """
        if hasattr(self, "last_run"):
            if self.last_run is None:
                return -1
            return self.last_run

        if not self.tasks.exists():
            return -1
        return self.tasks.all().aggregate(models.Max("run"))["run__max"]

    def is_final(self) -> bool:
        """
        Helper to tell whether a workflow is final.
        A workflow is considered final when it is in one of the final states:
        :attr:`~State.Completed`, :attr:`~State.Failed`, :attr:`~State.Error`, :attr:`~State.Stopped`

        :returns: Whether or not the workflow is considered final.
        """
        return self.state in FINAL_STATES

    @property
    def expiry(self):
        """
        A workflow's expiry date. This is the latest expiry date of its tasks.
        No action is taken when a workflow is expired.

        :returns: The latest expiry date of the workflow's tasks, or None if there are no tasks.
        :rtype: datetime or None
        """
        # This prevents performing another SQL request when tasks have already been prefetched.
        # See https://stackoverflow.com/a/19651840/5990435
        if (
            hasattr(self, "_prefetched_objects_cache")
            and self.tasks.field.remote_field.get_cache_name()
            in self._prefetched_objects_cache
        ):
            return max(t.expiry for t in self.tasks.all())
        else:
            return self.tasks.aggregate(models.Max("expiry"))["expiry__max"]

    def start(self):
        """
        Build new :class:`Task` instances associated to this workflow

        :raises AssertionError: If the workflow has already been started
        :returns: A dict mapping task slugs to task instances.
        :rtype: dict
        """
        assert (
            not self.tasks.exists()
        ), "Could not start the workflow as it has associated tasks"
        built_tasks = self.build_tasks()
        # setting last_run to 0 when a workflow gets started to avoid unnecessary db queries
        self.last_run = 0
        return built_tasks

    def retry(self):
        """
        Create new :class:`Task` instances with a new run number and resets the completion date.

        :raises AssertionError: If the workflow is not in a final state.
        :returns: A dict mapping task slugs to task instances.
        :rtype: dict
        """
        last_run = self.get_last_run()
        assert self.is_final()
        tasks = self.build_tasks(last_run + 1)
        # setting last_run so that subsequent calls to get_last_run do not require db queries
        self.last_run = last_run + 1
        self.finished = None
        self.save()
        return tasks

    def stop(self) -> None:
        """
        Fully stop the workflow by updating every running task to the :attr:`~State.Stopping` state,
        and every unscheduled task to the :attr:`~State.Stopped` state.
        """
        assert not self.is_final()
        stopping_count = Task.objects.filter(
            workflow=self, state__in=[State.Pending, State.Running]
        ).update(state=State.Stopping)
        Task.objects.filter(workflow=self, state=State.Unscheduled).update(
            state=State.Stopped
        )
        # If all the tasks are immediately stopped, then UpdateTask will not be able to update
        # the finished attribute, so we do it here.
        if not stopping_count:
            self.finished = timezone.now()
            self.save()


def expiry_default():
    """
    Default value for Task.expiry.

    :rtype: datetime
    """
    return timezone.now() + timedelta(days=30)


class Task(models.Model):
    """
    A task created from a workflow's recipe.
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    run = models.PositiveIntegerField()
    depth = models.PositiveIntegerField()
    # blank=False is only used in Django form validation, to prevent blank slugs we use a minimum length
    slug = models.CharField(max_length=250, validators=[MinLengthValidator(1)])
    priority = models.PositiveIntegerField(default=10)
    state = EnumField(State, default=State.Unscheduled, max_length=20)
    tags = CommaSeparatedListField(default=list, max_length=250)
    image = models.CharField(max_length=250)
    shm_size = models.CharField(max_length=80, blank=True, null=True, editable=False)
    command = models.TextField(blank=True, null=True)
    env = StringDictField(blank=True, null=True)
    has_docker_socket = models.BooleanField(default=False)
    image_artifact = models.ForeignKey(
        "ponos.Artifact",
        related_name="tasks_using_image",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    agent = models.ForeignKey(
        Agent,
        related_name="tasks",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    requires_gpu = models.BooleanField(default=False)
    gpu = models.OneToOneField(
        "ponos.GPU",
        related_name="task",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    workflow = models.ForeignKey(
        Workflow,
        related_name="tasks",
        on_delete=models.CASCADE,
    )
    parents = models.ManyToManyField(
        "self",
        related_name="children",
        symmetrical=False,
    )

    container = models.CharField(
        max_length=64,
        null=True,
        blank=True,
    )

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    expiry = models.DateTimeField(default=expiry_default)

    # Remote files required to start the container
    extra_files = StringDictField(default=dict)

    objects = TaskManager()

    class Meta:
        unique_together = (("workflow", "run", "slug"),)
        ordering = ("workflow", "run", "depth", "slug")

    def __str__(self) -> str:
        return "Task {}, run {}, depth {}".format(self.slug, self.run, self.depth)

    def get_absolute_url(self) -> str:
        """
        :returns: URL to the workflow details API for this workflow.
        """
        return reverse("ponos:task-details", args=[self.id])

    def is_final(self) -> bool:
        """
        Helper to tell whether a task is final.
        A task is considered final when it is in one of the final states:
        :attr:`~State.Completed`, :attr:`~State.Failed`, :attr:`~State.Error`, :attr:`~State.Stopped`

        :returns: Whether or not the workflow is considered final.
        """
        return self.state in FINAL_STATES

    @cached_property
    def s3_logs(self):
        """
        Get an S3 object instance for this task's logs file.
        This does not check for the file's existence.
        """
        return s3.Object(
            settings.PONOS_S3_LOGS_BUCKET,
            os.path.join(
                str(self.workflow.id),
                "run_{}".format(self.run),
                "{!s}.log".format(self.id),
            ),
        )

    @cached_property
    def s3_logs_get_url(self) -> str:
        """
        Get a presigned S3 GET URL for the task's logs file.
        This does not check for the file's existence.
        """
        return object_url("get_object", self.s3_logs)

    @cached_property
    def s3_logs_put_url(self) -> str:
        """
        Get a presigned S3 PUT URL for the task's logs file.
        This does not check for the file's existence.
        """
        return object_url("put_object", self.s3_logs)

    @property
    def has_logs(self) -> bool:
        """
        Test for the logs file's existence on S3.
        """
        try:
            self.s3_logs.load()
            return True
        except ClientError as e:
            if e.response["Error"]["Code"] != "404":
                raise
            return False

    @property
    def short_logs(self) -> str:
        """
        Fetch the last N bytes from the task's log file on S3, and return them decoded.
        An empty string is returned if the logs file is missing.
        The decoded text is cut to start after the first found newline character to avoid incomplete lines.

        The logs will start with ``[Logs were truncated]`` to help API users or frontends
        make it clear there are more logs available, unless less bytes than the requested amount are returned.

        The amount of fetched bytes is defined by the ``PONOS_LOG_TAIL`` setting, which defaults to 10000.
        """
        max_length = getattr(settings, "PONOS_LOG_TAIL", 10000)

        try:
            log_bytes = self.s3_logs.get(
                Range="bytes=-{}".format(max_length),
            )["Body"].read()

        except s3.meta.client.exceptions.NoSuchKey:
            return ""

        text = log_bytes.decode("utf-8", errors="replace")

        if len(text) < max_length:
            return text

        # Not starting from the beginning of the file
        # Start at the next new line
        text = text[text.find("\n") + 1 :]
        return "[Logs were truncated]\n" + text


def artifact_max_size():
    """
    AWS restricts uploads to 5GiB per PUT request,
    but some S3 implementations might not have this restriction,
    so we default to 5 GiB and allow overriding through the Django settings.

    We use a function to allow this setting to change at runtime,
    which makes unit testing a lot easier.
    """
    setting = getattr(settings, "PONOS_ARTIFACT_MAX_SIZE", None)
    return setting if setting is not None else 5 * 1024**3


class Artifact(models.Model):
    """
    A task Artifact (Json report, docker images, ML Models...)
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    task = models.ForeignKey(Task, related_name="artifacts", on_delete=models.CASCADE)
    path = models.CharField(max_length=500)
    size = models.BigIntegerField(
        validators=[
            MinValueValidator(1),
            MaxValueValidator(artifact_max_size),
        ]
    )
    content_type = models.CharField(max_length=250, default="application/octet-stream")
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ("task", "path")
        unique_together = (("task", "path"),)

    @cached_property
    def s3(self):
        """
        Get an S3 object instance for this Artifact
        This does not check for the file's existence.
        """
        return s3.Object(
            settings.PONOS_S3_ARTIFACTS_BUCKET,
            os.path.join(
                str(self.task.workflow_id),
                str(self.task.id),
                str(self.path),
            ),
        )

    @cached_property
    def s3_get_url(self) -> str:
        """
        Get a presigned S3 GET URL to download this artifact
        This does not check for the file's existence.
        """
        return object_url("get_object", self.s3)

    @cached_property
    def s3_put_url(self) -> str:
        """
        Get a presigned S3 PUT URL to store this Artifact on S3
        This does not check for the file's existence.
        """
        return object_url("put_object", self.s3)


def build_aes_cipher(nonce):
    """
    Initialize an AES cipher using the Ponos private key
    """
    key_path = getattr(settings, "PONOS_PRIVATE_KEY", None)
    assert key_path and os.path.exists(key_path), "Missing a PONOS_PRIVATE_KEY"
    # Use the Ponos private key SHA-256 as the AES key
    with open(settings.PONOS_PRIVATE_KEY, "rb") as f:
        ponos_key = f.read()
    aes_key = sha256(ponos_key).digest()
    return Cipher(algorithms.AES(aes_key), modes.CTR(nonce))


def encrypt(nonce, plain_text):
    """
    Encrypt a plain text using the AES cipher model content ciphering a text from the Ponos private key
    """
    cipher = build_aes_cipher(nonce)
    encryptor = cipher.encryptor()
    plain_text = plain_text.encode()
    return encryptor.update(plain_text)


class Secret(models.Model):
    """
    A secret encrypted with a derivate of the Ponos server private key (ECDH)
    Secret content is encrypted with AES, using a key generated serializing ponos private key (ECDH) with a salt
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    name = models.CharField(max_length=250, unique=True)
    nonce = models.BinaryField(max_length=16, default=gen_nonce)
    content = models.BinaryField(editable=True)

    def __str__(self):
        return self.name

    def decrypt(self):
        """
        Returns the plain text deciphered from the Ponos private key
        """
        decryptor = build_aes_cipher(self.nonce).decryptor()
        plain_text = decryptor.update(self.content)
        return plain_text.decode()


class GPU(models.Model):
    """
    A Graphic Card, attached to an agent
    """

    # The specific GPU uuid defined by the card itself
    # It's used as GPU-{id} to identity the GPU for docker
    id = models.UUIDField(primary_key=True)
    agent = models.ForeignKey(Agent, related_name="gpus", on_delete=models.CASCADE)
    name = models.CharField(max_length=250)

    # The numeric index assigned to the card by the host
    index = models.PositiveIntegerField()

    # RAM management
    ram_total = models.BigIntegerField(validators=[MinValueValidator(1)])

    class Meta:
        unique_together = (("agent_id", "index"),)

    def __str__(self):
        return self.name
