from rest_framework.permissions import SAFE_METHODS

from arkindex.ponos.models import Task
from arkindex.project.mixins import CorpusACLMixin
from arkindex.project.permissions import IsAuthenticated, require_internal


def require_agent(request, view):
    return getattr(request.user, 'is_admin', False) or getattr(request.user, 'is_agent', False)


def require_agent_or_internal(request, view):
    return require_internal(request, view) or getattr(request.user, 'is_agent', False)


class IsAgent(IsAuthenticated):
    """
    Only allow Ponos agents and admins.
    """
    checks = IsAuthenticated.checks + (require_agent, )


class IsAgentOrReadOnly(IsAgent):
    """
    Restricts write access to Ponos agents and admins,
    leave read access to everyone without any authentication.
    """
    allow_safe_methods = True


class IsAssignedAgentOrReadOnly(IsAgentOrReadOnly):
    """
    Restricts write access to Ponos agents and admins,
    and restricts write access on tasks to admins and agents assigned to them.
    Leave read access to everyone without any authentication.
    """

    def has_object_permission(self, request, view, obj) -> bool:
        if (
            isinstance(obj, Task)
            and not getattr(request.user, 'is_admin', False)
            and not request.user == obj.agent
        ):
            return request.method in SAFE_METHODS
        return super().has_object_permission(request, view, obj)


class IsAgentOrTaskAdmin(CorpusACLMixin, IsAuthenticated):
    """
    Permission to access a task with high privilege

    Allowed for admins, agents, internal users, creators of the process
    whose workflow contains the task, and users with an admin right on the
    process' corpus.
    """

    def has_object_permission(self, request, view, task):
        # Add request to attributes for the ACL mixin to work with self.user
        self.request = request

        return (
            require_agent(request, view)
            or require_internal(request, view)
            or (
                task.workflow.process is not None
                and task.workflow.process.corpus_id is not None
                and self.has_admin_access(task.workflow.process.corpus)
            )
        )


class IsAgentOrArtifactAdmin(IsAgentOrTaskAdmin):
    """
    Permission to access an artifact with high privilege, based on
    access to the artifact's task through IsAgentOrTaskAdmin.
    """

    def has_object_permission(self, request, view, artifact):
        return super().has_object_permission(request, view, artifact.task)


class IsAgentOrInternal(IsAuthenticated):
    """
    Allow access to agents or internal users, and not admins.
    """
    checks = (require_agent_or_internal, )
