from django.core import validators


class HiddenCallableValidatorMixin(object):
    """
    Implements a workaround for some issues with error messages in DRF
    and with drf-spectacular OpenAPI schema generation when the `limit_value`
    of any validator extending django.core.validators.BaseValidator is
    a callable.  This rewrites `self.limit_value` as a property,
    which calls the original limit value when it is callable while making
    Django, DRF and Spectacular believe it isn't callable.

    https://github.com/encode/django-rest-framework/discussions/8833
    https://github.com/tfranzel/drf-spectacular/issues/913
    """

    def __init__(self, limit_value, message=None):
        self._limit_value = limit_value
        if message:
            self.message = message

    @property
    def limit_value(self):
        return self._limit_value() if callable(self._limit_value) else self._limit_value


class MaxValueValidator(HiddenCallableValidatorMixin, validators.MaxValueValidator):
    pass
