from datetime import datetime
from unittest.mock import MagicMock, call, patch
from uuid import uuid4

from django.urls import reverse
from rest_framework import status
from rq.job import JobStatus

from arkindex.project.tests import FixtureAPITestCase


class MockedJob(object):

    def __init__(self, user_id=None, status=JobStatus.QUEUED):
        self.id = str(uuid4())
        self.enqueued_at = datetime(2020, 1, 1, 13, 37, 42)
        self.started_at = None
        self.ended_at = None
        self.progress = None
        self.description = 'something'
        self.user_id = user_id
        self._status = status

        self.delete = MagicMock()

    def get_status(self, refresh=True):
        # Those endpoints should not be setting refresh to True as they do not have to reload from Redis
        assert refresh is False, 'refresh should be set to False'
        return self._status


@patch('arkindex.project.rq_overrides.Queue.fetch_job')
@patch('arkindex.project.rq_overrides.UserRegistry.get_job_ids')
class TestJobs(FixtureAPITestCase):

    def setUp(self):
        super().setUp()
        self.user_mocked_job = MockedJob(self.user.id)
        self.superuser_mocked_job = MockedJob(self.superuser.id)

    def test_list_requires_login(self, ids_mock, fetch_mock):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:jobs-list'))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(ids_mock.called)
        self.assertFalse(fetch_mock.called)

    def test_list_requires_verified(self, ids_mock, fetch_mock):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:jobs-list'))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(ids_mock.called)
        self.assertFalse(fetch_mock.called)

    def test_list(self, ids_mock, fetch_mock):
        ids_mock.side_effect = [[self.user_mocked_job.id], []]
        fetch_mock.return_value = self.user_mocked_job

        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:jobs-list'))

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [
            {
                'id': self.user_mocked_job.id,
                'status': 'queued',
                'enqueued_at': '2020-01-01T13:37:42Z',
                'started_at': None,
                'ended_at': None,
                'progress': None,
                'description': 'something'
            }
        ])

        self.assertEqual(ids_mock.call_count, 2)
        self.assertEqual(fetch_mock.call_count, 1)

    def test_retrieve_requires_login(self, ids_mock, fetch_mock):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:jobs-retrieve', kwargs={'pk': self.user_mocked_job.id}))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(ids_mock.called)
        self.assertFalse(fetch_mock.called)

    def test_retrieve_requires_verified(self, ids_mock, fetch_mock):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:jobs-retrieve', kwargs={'pk': self.user_mocked_job.id}))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(ids_mock.called)
        self.assertFalse(fetch_mock.called)

    def test_retrieve(self, ids_mock, fetch_mock):
        fetch_mock.return_value = self.user_mocked_job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:jobs-retrieve', kwargs={'pk': self.user_mocked_job.id}))

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': self.user_mocked_job.id,
            'status': 'queued',
            'enqueued_at': '2020-01-01T13:37:42Z',
            'started_at': None,
            'ended_at': None,
            'progress': None,
            'description': 'something'
        })

        self.assertEqual(fetch_mock.call_count, 1)
        self.assertEqual(fetch_mock.call_args, call(self.user_mocked_job.id))

    def test_retrieve_not_found(self, ids_mock, fetch_mock):
        fetch_mock.return_value = None
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:jobs-retrieve', kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'})
            )

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        # Called once per queue
        self.assertEqual(fetch_mock.call_count, 2)
        self.assertListEqual(fetch_mock.call_args_list, [
            call('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'),
            call('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa')
        ])

    def test_retrieve_wrong_user(self, ids_mock, fetch_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:jobs-retrieve', kwargs={'pk': self.superuser_mocked_job.id}))

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_destroy_requires_login(self, ids_mock, fetch_mock):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:jobs-retrieve', kwargs={'pk': self.user_mocked_job.id}))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(ids_mock.called)
        self.assertFalse(fetch_mock.called)

    def test_destroy_requires_verified(self, ids_mock, fetch_mock):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(reverse('api:jobs-retrieve', kwargs={'pk': self.user_mocked_job.id}))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(ids_mock.called)
        self.assertFalse(fetch_mock.called)

    def test_destroy(self, ids_mock, fetch_mock):
        fetch_mock.return_value = self.user_mocked_job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(reverse('api:jobs-retrieve', kwargs={'pk': self.user_mocked_job.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(fetch_mock.call_count, 1)
        self.assertEqual(fetch_mock.call_args, call(self.user_mocked_job.id))
        self.assertEqual(self.user_mocked_job.delete.call_count, 1)

    def test_destroy_not_found(self, ids_mock, fetch_mock):
        fetch_mock.return_value = None
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(
                reverse('api:jobs-retrieve', kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'})
            )

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(fetch_mock.call_count, 2)
        self.assertListEqual(fetch_mock.call_args_list, [
            call('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'),
            call('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'),
        ])

    def test_destroy_wrong_user(self, ids_mock, fetch_mock):
        fetch_mock.return_value = self.superuser_mocked_job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:jobs-retrieve', kwargs={'pk': self.superuser_mocked_job.id}))

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(fetch_mock.call_count, 1)
        self.assertEqual(fetch_mock.call_args, call(self.superuser_mocked_job.id))

    def test_destroy_not_started(self, ids_mock, fetch_mock):
        started_job = MockedJob(user_id=self.user.id, status=JobStatus.STARTED)
        fetch_mock.return_value = started_job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(reverse('api:jobs-retrieve', kwargs={'pk': started_job.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ['Cannot delete a running job.'])

        self.assertEqual(fetch_mock.call_count, 1)
        self.assertEqual(fetch_mock.call_args, call(started_job.id))
        self.assertEqual(started_job.delete.call_count, 0)
